#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Sampler warnings."""

from collections import namedtuple
import enum

SamplerWarning = namedtuple("SamplerWarning", "kind, message, level, step, exec_info, extra")


@enum.unique
class WarningType(enum.Enum):
    """Enumeration of sampler warnings."""

    # For HMC and NUTS
    DIVERGENCE = 1
    TUNING_DIVERGENCE = 2
    DIVERGENCES = 3
    TREEDEPTH = 4
    # Problematic sampler parameters
    BAD_PARAMS = 5
    # Indications that chains did not converge, eg Rhat
    CONVERGENCE = 6
    BAD_ACCEPTANCE = 7
    BAD_ENERGY = 8



#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Hamiltonian Monte Carlo sampler."""

from typing import Callable, Tuple, Optional
import numpy as np

from .integration import IntegrationError
from .base_hmc import BaseHMC, HMCStepData, DivergenceInfo


__all__ = ["HamiltonianMC"]


class HamiltonianMC(BaseHMC):
    r"""A sampler for continuous variables based on Hamiltonian mechanics.

    See NUTS sampler for automatically tuned stopping time and step size
    scaling.
    """

    name = "hmc"
    generates_stats = True
    stats_dtypes = [
        {
            "step_size": np.float64,
            "n_steps": np.int64,
            "tune": np.bool,
            "step_size_bar": np.float64,
            "accept": np.float64,
            "diverging": np.bool,
            "energy_error": np.float64,
            "energy": np.float64,
            "path_length": np.float64,
            "accepted": np.bool,
            "model_logp": np.float64,
        }
    ]

    def __init__(
        self,
        logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
        model_ndim: int,
        scaling: Optional[np.ndarray] = None,
        is_cov: bool = False,
        potential=None,
        target_accept: float = 0.8,
        Emax: float = 1000,
        adapt_step_size: bool = True,
        step_scale: float = 0.25,
        gamma: float = 0.05,
        k: float = 0.75,
        t0: int = 10,
        step_rand: Optional[Callable[[float], float]] = None,
        path_length: float = 2.0,
        max_steps: int = 1024,
    ):
        """Set up the Hamiltonian Monte Carlo sampler.

        Parameters
        ----------
        logp_dlogp_func : Python callable
            Python callable that returns the log-probability and derivative of
            the log-probability, respectively.
        model_ndim : int
            Total number of parameters. Dimensionality of the output of
            ``logp_dlogp_func``.
        scaling : 1 or 2-dimensional array-like
            Scaling for momentum distribution. 1 dimensional arrays are
            interpreted as a matrix diagonal.
        is_cov : bool
            Treat scaling as a covariance matrix/vector if True, else treat
            it as a precision matrix/vector
        potential : littlemcmc.quadpotential.Potential, optional
            An object that represents the Hamiltonian with methods ``velocity``,
            ``energy``, and ``random`` methods. Only one of ``scaling`` or
            ``potential`` may be non-None.
        target_accept : float
            Adapt the step size such that the average acceptance probability
            across the trajectories are close to target_accept. Higher values
            for target_accept lead to smaller step sizes. Setting this to higher
            values like 0.9 or 0.99 can help with sampling from difficult
            posteriors. Valid values are between 0 and 1 (exclusive).
        Emax : float
            The maximum allowable change in the value of the Hamiltonian. Any
            trajectories that result in changes in the value of the Hamiltonian
            larger than ``Emax`` will be declared divergent.
        adapt_step_size : bool, default=True
            If True, performs dual averaging step size adaptation. If False,
            ``k``, ``t0``, ``gamma`` and ``target_accept`` are ignored.
        step_scale : float
            Size of steps to take, automatically scaled down by 1 / (``size`` **
            0.25).
        gamma : float, default .05
        k : float, default .75
            Parameter for dual averaging for step size adaptation. Values
            between 0.5 and 1 (exclusive) are admissible. Higher values
            correspond to slower adaptation.
        t0 : int, default 10
            Parameter for dual averaging. Higher values slow initial adaptation.
        step_rand : Python callable
            Callback for step size adaptation. Called on the step size at each
            iteration immediately before performing dual-averaging step size
            adaptation.
        path_length : float, default=2
            total length to travel
        max_steps : int, default=1024
            The maximum number of leapfrog steps.
        """
        super(HamiltonianMC, self).__init__(
            scaling=scaling,
            step_scale=step_scale,
            is_cov=is_cov,
            logp_dlogp_func=logp_dlogp_func,
            model_ndim=model_ndim,
            potential=potential,
            Emax=Emax,
            target_accept=target_accept,
            gamma=gamma,
            k=k,
            t0=t0,
            adapt_step_size=adapt_step_size,
            step_rand=step_rand,
        )
        self.path_length = path_length
        self.max_steps = max_steps

    def _hamiltonian_step(self, start: np.ndarray, p0: np.ndarray, step_size: float) -> HMCStepData:
        path_length = np.random.rand() * self.path_length
        n_steps = max(1, int(path_length / step_size))
        n_steps = min(self.max_steps, n_steps)

        energy_change = -np.inf
        state = start
        div_info = None
        try:
            for _ in range(n_steps):
                state = self.integrator.step(step_size, state)
        except IntegrationError as e:
            div_info = DivergenceInfo("Divergence encountered.", e, state)
        else:
            if not np.isfinite(state.energy):
                div_info = DivergenceInfo("Divergence encountered, bad energy.", None, state)
            energy_change = start.energy - state.energy
            if np.isnan(energy_change):
                energy_change = -np.inf
            if np.abs(energy_change) > self.Emax:
                div_info = DivergenceInfo(
                    "Divergence encountered, large integration error.", None, state
                )

        accept_stat = min(1, np.exp(energy_change))

        if div_info is not None or np.random.rand() >= accept_stat:
            end = start
            accepted = False
        else:
            end = state
            accepted = True

        stats = {
            "path_length": path_length,
            "n_steps": n_steps,
            "accept": accept_stat,
            "energy_error": energy_change,
            "energy": state.energy,
            "accepted": accepted,
            "model_logp": state.model_logp,
        }
        return HMCStepData(end, accept_stat, div_info, stats)



#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Leapfrog integrators."""

from collections import namedtuple
from typing import Callable, Tuple

import numpy as np
from scipy import linalg
from .quadpotential import QuadPotential


State = namedtuple("State", "q, p, v, q_grad, energy, model_logp")


class IntegrationError(RuntimeError):
    """Numerical errors during leapfrog integration."""

    pass


class CpuLeapfrogIntegrator(object):
    """Leapfrog integrator using the CPU."""

    def __init__(
        self,
        potential: QuadPotential,
        logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
    ) -> None:
        """Instantiate a CPU leapfrog integrator.

        Parameters
        ----------
        potential
        logp_dlogp_func
        """
        self._potential = potential
        self._logp_dlogp_func = logp_dlogp_func

    def compute_state(self, q: np.ndarray, p: np.ndarray) -> State:
        """Compute Hamiltonian functions using a position and momentum.

        Parameters
        ----------
        q
            Position.
        p
            Momentum
        """
        logp, dlogp = self._logp_dlogp_func(q)
        v = self._potential.velocity(p)
        kinetic = self._potential.energy(p, velocity=v)
        energy = kinetic - logp
        return State(q, p, v, dlogp, energy, logp)

    def step(self, epsilon, state: State, out=None):
        """Leapfrog integrator step.

        Half a momentum update, full position update, half momentum update.

        Parameters
        ----------
        epsilon: float, > 0
            step scale
        state: State namedtuple,
            current position data
        out: (optional) State namedtuple,
            preallocated arrays to write to in place

        Returns
        -------
        None if `out` is provided, else a State namedtuple
        """
        try:
            return self._step(epsilon, state)
        except linalg.LinAlgError as err:
            msg = "LinAlgError during leapfrog step."
            raise IntegrationError(msg)
        except ValueError as err:
            # Raised by many scipy.linalg functions
            scipy_msg = "array must not contain infs or nans"
            if len(err.args) > 0 and scipy_msg in err.args[0].lower():
                msg = "Infs or nans in scipy.linalg during leapfrog step."
                raise IntegrationError(msg)
            else:
                raise

    def _step(self, epsilon, state: State) -> State:
        """Perform one leapfrog step."""
        pot = self._potential
        q, p, v, q_grad, energy, logp = state

        dt = 0.5 * epsilon

        # Half momentum step
        p_new = p + dt * q_grad

        # Whole position step
        v_new = pot.velocity(p_new)
        q_new = (q + epsilon * v_new).astype(q.dtype)

        # Half momentum step
        logp, q_new_grad = self._logp_dlogp_func(q_new)
        p_new = p_new + dt * q_new_grad

        kinetic = pot.velocity_energy(p_new, v_new)
        energy = kinetic - logp

        return State(q_new, p_new, v_new, q_new_grad, energy, logp)



#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Utility mathematical functions."""

import numpy as np
import numpy.random as nr


def logbern(log_p: float) -> bool:
    """Perform a Bernoulli trial given a log probability."""
    if np.isnan(log_p):
        raise FloatingPointError("log_p can't be nan.")
    return np.log(nr.uniform()) < log_p


def log1mexp_numpy(x: float) -> float:
    """
    Compute log(1 - exp(-x)).

    This function is numerically more stable than the naive approach. For details, see
    https://cran.r-project.org/web/packages/Rmpfr/vignettes/log1mexp-note.pdf
    """
    return np.where(x < 0.683, np.log(-np.expm1(-x)), np.log1p(-np.exp(-x)))


def logdiffexp_numpy(a: float, b: float) -> float:
    """Compute log(exp(a) - exp(b))."""
    return a + log1mexp_numpy(a - b)



#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Dual averaging step size adaptation."""

import numpy as np
from scipy import stats

from .report import SamplerWarning, WarningType


class DualAverageAdaptation(object):
    """Dual averaging step size adaptation."""

    def __init__(self, initial_step, target, gamma, k, t0):
        """Class for dual averaging step size adaptation.

        Parameters
        ----------
        initial_step
        target
        gamma : float, default .05
        k : float, default .75
            Parameter for dual averaging for step size adaptation. Values
            between 0.5 and 1 (exclusive) are admissible. Higher values
            correspond to slower adaptation.
        t0 : int, default 10
            Parameter for dual averaging. Higher values slow initial
            adaptation.
        """
        self._initial_step = initial_step
        self._target = target
        self._k = k
        self._t0 = t0
        self._gamma = gamma
        self.reset()

    def reset(self):
        """Reset step size adaptation routine."""
        self._log_step = np.log(self._initial_step)
        self._log_bar = self._log_step
        self._hbar = 0.0
        self._count = 1
        self._mu = np.log(10 * self._initial_step)
        self._tuned_stats = []

    def current(self, tune):
        """Get current step size.

        Parameters
        ----------
        tune : bool
            True during tuning, else False.
        """
        if tune:
            return np.exp(self._log_step)
        else:
            return np.exp(self._log_bar)

    def update(self, accept_stat, tune):
        """Update step size.

        Parameters
        ----------
        accept_stat : float
            HMC step acceptance statistic.
        tune : bool
            True during tuning, else False.
        """
        if not tune:
            self._tuned_stats.append(accept_stat)
            return

        count, k, t0 = self._count, self._k, self._t0
        w = 1.0 / (count + t0)
        self._hbar = (1 - w) * self._hbar + w * (self._target - accept_stat)

        self._log_step = self._mu - self._hbar * np.sqrt(count) / self._gamma
        mk = count ** -k
        self._log_bar = mk * self._log_step + (1 - mk) * self._log_bar
        self._count += 1

    def stats(self):
        """Get step size adaptation statistics."""
        return {
            "step_size": np.exp(self._log_step),
            "step_size_bar": np.exp(self._log_bar),
        }

    def warnings(self):
        """Generate warnings from dual averaging step size adaptation."""
        accept = np.array(self._tuned_stats)
        mean_accept = np.mean(accept)
        target_accept = self._target
        # Try to find a reasonable interval for acceptable acceptance
        # probabilities. Finding this was mostly trial and error.
        n_bound = min(100, len(accept))
        n_good, n_bad = mean_accept * n_bound, (1 - mean_accept) * n_bound
        lower, upper = stats.beta(n_good + 1, n_bad + 1).interval(0.95)
        if target_accept < lower or target_accept > upper:
            msg = (
                "The acceptance probability does not match the target. It "
                "is %s, but should be close to %s. Try to increase the "
                "number of tuning steps." % (mean_accept, target_accept)
            )
            info = {"target": target_accept, "actual": mean_accept}
            warning = SamplerWarning(WarningType.BAD_ACCEPTANCE, msg, "warn", None, None, info)
            return [warning]
        else:
            return []



#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Base class for Hamiltonian Monte Carlo samplers."""

from collections import namedtuple
from typing import Callable, Tuple, List, Optional
import numpy as np

from . import integration, step_sizes
from .quadpotential import quad_potential, QuadPotential, QuadPotentialDiagAdapt
from .report import SamplerWarning, WarningType

HMCStepData = namedtuple("HMCStepData", "end, accept_stat, divergence_info, stats")
DivergenceInfo = namedtuple("DivergenceInfo", "message, exec_info, state")


class BaseHMC:
    """Superclass to implement Hamiltonian Monte Carlo."""

    def __init__(
        self,
        logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
        model_ndim: int,
        scaling: Optional[np.ndarray],
        is_cov: bool,
        potential: QuadPotential,
        target_accept: float,
        Emax: float,
        adapt_step_size: bool,
        step_scale: float,
        gamma: float,
        k: float,
        t0: int,
        step_rand: Optional[Callable[[float], float]],
    ) -> None:
        """Set up Hamiltonian samplers with common structures.

        Parameters
        ----------
        logp_dlogp_func : Python callable
            Python callable that returns the log-probability and derivative of
            the log-probability, respectively.
        model_ndim : int
            Total number of parameters. Dimensionality of the output of
            ``logp_dlogp_func``.
        scaling : 1 or 2-dimensional array-like
            Scaling for momentum distribution. 1 dimensional arrays are
            interpreted as a matrix diagonal. Only one of ``scaling`` or
            ``potential`` may be non-None.
        is_cov : bool
            Treat scaling as a covariance matrix/vector if True, else treat
            it as a precision matrix/vector
        potential : littlemcmc.quadpotential.Potential, optional
            An object that represents the Hamiltonian with methods ``velocity``,
            ``energy``, and ``random`` methods. Only one of ``scaling`` or
            ``potential`` may be non-None.
        target_accept : float
            Adapt the step size such that the average acceptance probability
            across the trajectories are close to target_accept. Higher values
            for target_accept lead to smaller step sizes. Setting this to higher
            values like 0.9 or 0.99 can help with sampling from difficult
            posteriors. Valid values are between 0 and 1 (exclusive).
        Emax : float
            The maximum allowable change in the value of the Hamiltonian. Any
            trajectories that result in changes in the value of the Hamiltonian
            larger than ``Emax`` will be declared divergent.
        adapt_step_size : bool, default=True
            If True, performs dual averaging step size adaptation. If False,
            ``k``, ``t0``, ``gamma`` and ``target_accept`` are ignored.
        step_scale : float
            Size of steps to take, automatically scaled down by 1 / (``size`` **
            0.25).
        gamma : float, default .05
        k : float, default .75
            Parameter for dual averaging for step size adaptation. Values
            between 0.5 and 1 (exclusive) are admissible. Higher values
            correspond to slower adaptation.
        t0 : int, default 10
            Parameter for dual averaging. Higher values slow initial adaptation.
        step_rand : Python callable
            Callback for step size adaptation. Called on the step size at each
            iteration immediately before performing dual-averaging step size
            adaptation.
        """
        self._logp_dlogp_func = logp_dlogp_func
        self.adapt_step_size = adapt_step_size
        self.Emax = Emax
        self.iter_count = 0
        self.model_ndim = model_ndim
        self.step_size = step_scale / (model_ndim ** 0.25)
        self.target_accept = target_accept
        self.step_adapt = step_sizes.DualAverageAdaptation(
            self.step_size, target_accept, gamma, k, t0
        )
        self.tune = True

        if scaling is None and potential is None:
            # Default to diagonal quadpotential
            mean = np.zeros(model_ndim)
            var = np.ones(model_ndim)
            potential = QuadPotentialDiagAdapt(model_ndim, mean, var, 10)

        if scaling is not None and potential is not None:
            raise ValueError("Cannot specify both `potential` and `scaling`.")
        elif potential is not None:
            self.potential = potential
        else:
            self.potential = quad_potential(scaling, is_cov)

        self.integrator = integration.CpuLeapfrogIntegrator(self.potential, self._logp_dlogp_func)
        self._step_rand = step_rand
        self._warnings: List[SamplerWarning] = []
        self._samples_after_tune = 0
        self._num_divs_sample = 0

    def stop_tuning(self) -> None:
        """Stop tuning."""
        if hasattr(self, "tune"):
            self.tune = False

    def _hamiltonian_step(self, start: np.ndarray, p0: np.ndarray, step_size: float) -> HMCStepData:
        """Compute one Hamiltonian trajectory and return the next state.

        Subclasses must overwrite this method and return a `HMCStepData`.
        """
        raise NotImplementedError("Abstract method")

    def _astep(self, q0: np.ndarray):
        """Perform a single HMC iteration."""
        p0 = self.potential.random()
        start = self.integrator.compute_state(q0, p0)

        if not np.isfinite(start.energy):
            raise ValueError(
                "Bad initial energy: {}. The model might be misspecified.".format(start.energy)
            )

        # Adapt step size
        adapt_step = self.tune and self.adapt_step_size
        step_size = self.step_adapt.current(adapt_step)
        self.step_size = step_size
        if self._step_rand is not None:
            step_size = self._step_rand(step_size)

        # Take the Hamiltonian step
        hmc_step = self._hamiltonian_step(start, p0, step_size)

        # Update step size and quadpotential
        self.step_adapt.update(hmc_step.accept_stat, adapt_step)
        self.potential.update(hmc_step.end.q, hmc_step.end.q_grad, self.tune)

        if hmc_step.divergence_info:
            info = hmc_step.divergence_info
            if self.tune:
                kind = WarningType.TUNING_DIVERGENCE
                point = None
            else:
                kind = WarningType.DIVERGENCE
                self._num_divs_sample += 1
                # We don't want to fill up all memory, so do not return points
                # with divergence info
                point = None
            warning = SamplerWarning(
                kind, info.message, "debug", self.iter_count, info.exec_info, point
            )

            self._warnings.append(warning)

        self.iter_count += 1
        if not self.tune:
            self._samples_after_tune += 1

        stats = {"tune": self.tune, "diverging": bool(hmc_step.divergence_info)}

        stats.update(hmc_step.stats)
        stats.update(self.step_adapt.stats())

        return hmc_step.end.q, [stats]

    def reset_tuning(self, start: np.ndarray = None) -> None:
        """Reset quadpotential and step size adaptation, and begin retuning."""
        self.step_adapt.reset()
        self.reset(start=None)

    def reset(self, start: np.ndarray = None) -> None:
        """Reset quadpotential and begin retuning."""
        self.tune = True
        self.potential.reset()

    def warnings(self) -> List[SamplerWarning]:
        """Generate warnings from HMC sampler."""
        # list.copy() is only available in Python 3
        warnings = self._warnings.copy()

        # Generate a global warning for divergences
        message = ""
        n_divs = self._num_divs_sample
        if n_divs and self._samples_after_tune == n_divs:
            message = (
                "The chain contains only diverging samples. The model is probably misspecified."
            )
        elif n_divs == 1:
            message = (
                "There was 1 divergence after tuning. Increase "
                "`target_accept` or reparameterize."
            )
        elif n_divs > 1:
            message = (
                "There were %s divergences after tuning. Increase "
                "`target_accept` or reparameterize." % n_divs
            )

        if message:
            warning = SamplerWarning(WarningType.DIVERGENCES, message, "error", None, None, None)
            warnings.append(warning)

        warnings.extend(self.step_adapt.warnings())
        return warnings



import torch
import littlemcmc as lmc
import numpy as np


np.random.seed(42)

true_params = np.array([0.5, -2.3, -0.23])

N = 50
t = np.linspace(0, 10, 2)
x = np.random.uniform(0, 10, 50)
y = x * true_params[0] + true_params[1]
y_obs = y + np.exp(true_params[-1]) * np.random.randn(N)


class LinearModel(torch.nn.Module):
    def __init__(self):
        super(LinearModel, self).__init__()
        self.m = torch.nn.Parameter(torch.tensor(0.0, dtype=torch.float64))
        self.b = torch.nn.Parameter(torch.tensor(0.0, dtype=torch.float64))
        self.logs = torch.nn.Parameter(torch.tensor(0.0, dtype=torch.float64))

    def forward(self, x, y):
        mean = self.m * x + self.b
        loglike = -0.5 * torch.sum((y - mean) ** 2 * torch.exp(-2 * self.logs) + 2 * self.logs)
        return loglike


torch_model = torch.jit.script(LinearModel())
torch_params = [torch_model.m, torch_model.b, torch_model.logs]
args = [torch.tensor(x, dtype=torch.double), torch.tensor(y_obs, dtype=torch.double)]


def torch_logp_dlogp_func(x):
    for i, p in enumerate(torch_params):
        p.data = torch.tensor(x[i])
        if p.grad is not None:
            p.grad.detach_()
            p.grad.zero_()

    result = torch_model(*args)
    result.backward()

    return result.detach().numpy(), np.array([p.grad.numpy() for p in torch_params])


trace, stats = lmc.sample(
    logp_dlogp_func=torch_logp_dlogp_func, model_ndim=3, tune=500, draws=1000, chains=4,
)



#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Sampling driver functions (unrelated to PyMC3's `sampling.py`)."""

import os
from collections.abc import Iterable
import pickle
from typing import Callable, Tuple, Optional, Union, List
import logging
import numpy as np
from fastprogress.fastprogress import progress_bar

from . import parallel_sampling as ps
from .nuts import NUTS
from .hmc import HamiltonianMC
from .quadpotential import QuadPotential, QuadPotentialDiagAdapt, QuadPotentialFullAdapt
from .report import SamplerWarning


_log = logging.getLogger("littlemcmc")


def sample(
    logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
    model_ndim: int,
    draws: int = 1000,
    tune: int = 1000,
    step: Union[NUTS, HamiltonianMC] = None,
    init: str = "auto",
    chains: Optional[int] = None,
    cores: Optional[int] = None,
    start: Optional[np.ndarray] = None,
    progressbar: Union[bool, str] = True,
    random_seed: Optional[Union[int, List[int]]] = None,
    discard_tuned_samples: bool = True,
    chain_idx: int = 0,
    callback=None,
    mp_ctx=None,
    pickle_backend: str = "pickle",
    **kwargs,
):
    """
    Draw samples from the posterior using the given step methods.

    Parameters
    ----------
    logp_dlogp_func: Python callable
        Python callable that returns a tuple of the model joint log probability and its
        derivative, in that order.
    model_ndim: int
        The number of parameters of the model.
    draws: int
        The number of samples to draw. Defaults to 1000. The number of tuned samples are
        discarded by default. See ``discard_tuned_samples``.
    tune: int
        Number of iterations to tune, defaults to 1000. Samplers adjust the step sizes,
        scalings or similar during tuning. Tuning samples will be drawn in addition to
        the number specified in the ``draws`` argument, and will be discarded unless
        ``discard_tuned_samples`` is set to False.
    step: function
        A step function. By default the NUTS step method will be used.
    init: str
        Initialization method to use for auto-assigned NUTS samplers.
            * auto: Choose a default initialization method automatically. Currently,
              this is ``jitter+adapt_diag``, but this can change in the future. If you
              depend on the exact behaviour, choose an initialization method explicitly.
            * adapt_diag: Start with a identity mass matrix and then adapt a diagonal
              based on the variance of the tuning samples.
            * jitter+adapt_diag: Same as ``adapt_diag``, but add uniform jitter in
              [-1, 1] to the starting point in each chain.
            * adapt_full: Same as `'adapt_diag'`, but adapt a dense mass matrix using
              the sample covariances.
    chains: int
        The number of chains to sample. Running independent chains is important for some
        convergence statistics and can also reveal multiple modes in the posterior. If
        ``None``, then set to either ``cores`` or 2, whichever is larger.
    cores: int
        The number of chains to run in parallel. If ``None``, set to the number of CPUs
        in the system, but at most 4.
    start: dict, or array of dict
        Starting point in parameter space. Initialization methods for NUTS (see ``init``
        keyword) can overwrite the default.
    progressbar: bool, optional default=True
        Whether or not to display a progress bar in the command line. The bar shows the
        percentage of completion, the sampling speed in samples per second (SPS), and
        the estimated remaining time until completion ("expected time of arrival"; ETA).
    random_seed: int or list of ints
        A list is accepted if ``cores`` is greater than one.
    discard_tuned_samples: bool
        Whether to discard posterior samples of the tune interval.

    Returns
    -------
    trace: np.array
        An array that contains the samples.
    stats: dict
        A dictionary that contains sampler statistics.

    Notes
    -----
    Optional keyword arguments can be passed to ``sample`` to be delivered to the
    ``step_method``s used during sampling. In particular, the NUTS step method accepts a
    number of arguments. You can find a full list of arguments in the docstring of the
    step methods. Common options are:
        * target_accept: float in [0, 1]. The step size is tuned such that we
          approximate this acceptance rate. Higher values like 0.9 or 0.95 often work
          better for problematic posteriors.
        * max_treedepth: The maximum depth of the trajectory tree.
        * step_scale: float, default 0.25. The initial guess for the step size scaled
        down by :math:`1/n**(1/4)`.
    """
    if cores is None:
        cores = min(4, os.cpu_count())
        assert cores is not None  # To make mypy happy
    if chains is None:
        chains = max(2, cores)
        assert chains is not None  # To make mypy happy

    if random_seed is None or isinstance(random_seed, int):
        if random_seed is not None:
            np.random.seed(random_seed)
        random_seed = [np.random.randint(2 ** 30) for _ in range(chains)]  # type: ignore
    elif isinstance(random_seed, Iterable) and len(random_seed) != chains:
        random_seed = random_seed[:chains]
    elif not isinstance(random_seed, Iterable):
        raise TypeError("Invalid value for `random_seed`. Must be tuple, list or int")

    # Small trace warning
    if draws == 0:
        msg = "Tuning was enabled throughout the whole trace."
        _log.warning(msg)
    elif draws < 500:
        msg = "Only {} samples in chain.".format(draws)
        _log.warning(msg)

    if step is None or start is None:
        start_, step_ = init_nuts(
            logp_dlogp_func=logp_dlogp_func,
            model_ndim=model_ndim,
            init=init,
            random_seed=random_seed,
            **kwargs,
        )
        if step is None:
            step = step_
        if start is None:
            start = start_

    if start is None:
        start = {}
    if isinstance(start, np.ndarray) or isinstance(start, dict):
        start = [start] * chains

    sample_args = {
        "logp_dlogp_func": logp_dlogp_func,
        "model_ndim": model_ndim,
        "draws": draws,
        "tune": tune,
        "step": step,
        "start": start,
        "chain": chain_idx,
        "chains": chains,
        "progressbar": progressbar,
        "random_seed": random_seed,
        "cores": cores,
        "callback": callback,
        "discard_tuned_samples": discard_tuned_samples,
    }
    parallel_args = {
        "pickle_backend": pickle_backend,
        "mp_ctx": mp_ctx,
    }

    parallel = cores > 1 and chains > 1
    if parallel:
        _log.info("Multiprocess sampling ({} chains in {} jobs)".format(chains, cores))
        try:
            traces, stats = _mp_sample(**sample_args, **parallel_args)  # type: ignore
        except pickle.PickleError:
            _log.warning("Could not pickle model, sampling singlethreaded.")
            _log.debug("Pickling error:", exec_info=True)
            parallel = False
        except AttributeError as e:
            if str(e).startswith("AttributeError: Can't pickle"):
                _log.warning("Could not pickle model, sampling singlethreaded.")
                _log.debug("Pickling error:", exec_info=True)
                parallel = False
            else:
                raise

    if not parallel:
        _log.info("Sequential sampling ({} chains in 1 job)".format(chains))
        traces, stats = _sample_many(**sample_args)  # type: ignore

    # Reshape `trace` to have shape [num_chains, num_samples, num_variables]
    trace = np.array([np.atleast_2d(chain_trace).T for chain_trace in traces])

    # Reshape `stats` to a dictionary with keys = string of sampling stat name,
    # values = np.array with shape [num_chains, num_samples, num_variables]
    stats = {
        name: np.array(
            [
                [np.atleast_1d(iter_stats[name]) for iter_stats in chain_stats]
                for chain_stats in stats
            ]
        ).astype(dtype)
        for (name, dtype) in step.stats_dtypes[0].items()
    }

    return trace, stats


def _mp_sample(
    model_ndim: int,
    draws: int,
    tune: int,
    step,
    chains: int,
    cores: int,
    chain: int,
    random_seed: list,
    start: list,
    progressbar=True,
    trace=None,
    callback=None,
    discard_tuned_samples=True,
    mp_ctx=None,
    pickle_backend="pickle",
    **kwargs,
):
    """
    Sample multiple chains in multiple processes.

    Main iteration for multiprocess sampling.

    Parameters
    ----------
    draws : int
        The number of samples to draw
    tune : int, optional
        Number of iterations to tune, if applicable (defaults to None)
    step : function
        Step function
    chains : int
        The number of chains to sample.
    cores : int
        The number of chains to run in parallel.
    chain : int
        Number of the first chain.
    random_seed : list of ints
        Random seeds for each chain.
    start : list
        Starting points for each chain.
    progressbar : bool
        Whether or not to display a progress bar in the command line.
    trace : backend, list, MultiTrace or None
        This should be a backend instance, a list of variables to track, or a MultiTrace object
        with past values. If a MultiTrace object is given, it must contain samples for the chain
        number ``chain``. If None or a list of variables, the NDArray backend is used.
    callback : Callable
        A function which gets called for every sample from the trace of a chain. The function is
        called with the trace and the current draw and will contain all samples for a single trace.
        the ``draw.chain`` argument can be used to determine which of the active chains the sample
        is drawn from.
        Sampling can be interrupted by throwing a ``KeyboardInterrupt`` in the callback.

    Returns
    -------
    trace : pymc3.backends.base.MultiTrace
        A ``MultiTrace`` object that contains the samples for all chains.
    """
    trace = np.zeros([chains, model_ndim, tune + draws])
    stats: List[List[SamplerWarning]] = [[] for _ in range(chains)]

    sampler = ps.ParallelSampler(
        model_ndim,
        draws,
        tune,
        chains,
        cores,
        random_seed,
        start,
        step,
        chain,
        progressbar,
        mp_ctx=mp_ctx,
        pickle_backend=pickle_backend,
    )
    try:
        try:
            with sampler:
                for draw in sampler:
                    trace[draw.chain, :, draw.draw_idx] = draw.point
                    stats[draw.chain].append(draw.stats[0])
                    if callback is not None:
                        callback(trace=trace, draw=draw)

        except ps.ParallelSamplingError as error:
            # trace = traces[error._chain - chain]
            # trace._add_warnings(error._warnings)
            # for trace in traces:
            #     trace.close()

            # multitrace = MultiTrace(traces)
            # multitrace._report._log_summary()
            raise
        if discard_tuned_samples and trace is not None and stats is not None:
            # pylint: disable=W0631
            trace = trace[:, :, tune:]
            stats = [chain_stats[tune:] for chain_stats in stats]
        return trace, stats
    except KeyboardInterrupt:
        if discard_tuned_samples and trace is not None and stats is not None:
            # pylint: disable=W0631
            trace = trace[:, :, tune:]
            stats = [chain_stats[tune:] for chain_stats in stats]


def _sample_many(
    logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
    model_ndim: int,
    draws: int,
    tune: int,
    chain: int,
    chains: int,
    start: list,
    random_seed: list,
    step,
    discard_tuned_samples=True,
    callback=None,
    **kwargs,
):
    """
    Sample all chains sequentially.

    Parameters
    ----------
    draws: int
        The number of samples to draw
    chain: int
        Number of the first chain in the sequence.
    chains: int
        Total number of chains to sample.
    start: list
        Starting points for each chain
    random_seed: list
        A list of seeds, one for each chain
    step: function
        Step function

    Returns
    -------
    trace, stats
    """
    traces: List[np.ndarray] = []
    stats = []

    for i in range(chains):
        trace, stats_ = _sample(
            logp_dlogp_func=logp_dlogp_func,
            model_ndim=model_ndim,
            draws=draws,
            tune=tune,
            random_seed=random_seed[i],
            chain=chain + i,
            start=start[i],
            step=step,
            discard_tuned_samples=discard_tuned_samples,
            callback=callback,
            **kwargs,
        )
        if trace is None:
            if len(traces) == 0:
                raise ValueError("Sampling stopped before a sample was created.")
            else:
                break
        # TODO: bring this back for easy keyboard interrupts...?
        # elif len(trace) < tune + draws:
        #     if len(traces) == 0:
        #         traces.append(trace)
        #         stats.append(stats_)
        #     break
        else:
            traces.append(trace)
            stats.append(stats_)

    return traces, stats


def _sample(
    logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
    model_ndim: int,
    chain: int,
    progressbar: bool,
    random_seed,
    start,
    draws: int,
    step=None,
    trace=None,
    tune=None,
    discard_tuned_samples=True,
    callback=None,
    **kwargs,
):
    """
    Sample one chain in one process.

    Parameters
    ----------
    logp_dlogp_func : Python callable
    model_ndim : int
    chain : int
        Number of the chain that the samples will belong to.
    progressbar : bool
        Whether or not to display a progress bar in the command line. The bar shows the percentage
        of completion, the sampling speed in samples per second (SPS), and the estimated remaining
        time until completion ("expected time of arrival"; ETA).
    random_seed : int or list of ints
        A list is accepted if ``cores`` is greater than one.
    start : dict
        Starting point in parameter space (or partial point)
    draws : int
        The number of samples to draw
    step : function
        Step function
    trace : backend, list, or MultiTrace
        This should be a backend instance, a list of variables to track, or a MultiTrace object
        with past values. If a MultiTrace object is given, it must contain samples for the chain
        number ``chain``. If None or a list of variables, the NDArray backend is used.
    tune : int, optional
        Number of iterations to tune, if applicable (defaults to None)
    callback : Python callable, optional

    Returns
    -------
    trace, stats
    """
    skip_first = kwargs.get("skip_first", 0)

    sampling = _iter_sample(
        logp_dlogp_func, model_ndim, draws, tune, step, start, random_seed, callback
    )
    _pbar_data = {"chain": chain, "divergences": 0}
    _desc = "Sampling chain {chain:d}, {divergences:,d} divergences"
    if progressbar:
        sampling = progress_bar(sampling, total=tune + draws, display=progressbar)
        sampling.comment = _desc.format(**_pbar_data)

    trace = None
    stats = None
    try:
        for it, (trace, stats) in enumerate(sampling):
            if it >= skip_first:  # and diverging:
                # FIXME: surface num divergences from `stats`
                # _pbar_data["divergences"] += 1
                if progressbar:
                    sampling.comment = _desc.format(**_pbar_data)
    except KeyboardInterrupt:
        pass

    if discard_tuned_samples and trace is not None and stats is not None:
        # pylint: disable=W0631
        trace = trace[:, tune:]
        stats = stats[tune:]

    return trace, stats


def _iter_sample(
    logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
    model_ndim: int,
    draws: int,
    tune: int,
    step: Union[NUTS, HamiltonianMC],
    start: np.ndarray,
    random_seed: Union[None, int, List[int]] = None,
    callback=None,
):
    """
    Yield one chain in one process.

    Main iterator for singleprocess sampling.
    """
    if random_seed is not None:
        np.random.seed(random_seed)

    q = start
    trace = np.zeros([model_ndim, tune + draws])
    stats: List[SamplerWarning] = []

    step.tune = bool(tune)
    if hasattr(step, "reset_tuning"):
        step.reset_tuning()

    for i in range(tune + draws):
        if i == 0 and hasattr(step, "iter_count"):
            step.iter_count = 0
        if i == tune:
            step.stop_tuning()
        q, step_stats = step._astep(q)
        trace[:, i] = q
        stats.extend(step_stats)
        if callback is not None:
            warns = getattr(step, "warnings", None)
            # FIXME: implement callbacks
            # callback(
            #     trace=trace, draw=(chain, i == draws, i, i < tune, stats, point, warns),
            # )
        yield trace, stats


def init_nuts(
    logp_dlogp_func: Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]],
    model_ndim: int,
    init: str = "auto",
    random_seed: Union[None, int, List[int]] = None,
    **kwargs,
):
    """Set up the mass matrix initialization for NUTS.

    NUTS convergence and sampling speed is extremely dependent on the
    choice of mass/scaling matrix. This function implements different
    methods for choosing or adapting the mass matrix.

    Parameters
    ----------
    init: str
        Initialization method to use.
        * auto : Choose a default initialization method automatically.
          Currently, this is `'jitter+adapt_diag'`, but this can change in the
          future. If you depend on the exact behaviour, choose an initialization
          method explicitly.
        * adapt_diag : Start with a identity mass matrix and then adapt a diagonal based
          on the variance of the tuning samples.
        * jitter+adapt_diag : Same as `'adapt_diag'`, but use uniform jitter in [-1, 1]
          as starting point in each chain.
        * adapt_full: Same as `'adapt_diag'`, but adapts a dense mass matrix using the
          sample covariances.
        * jitter+adapt_full: Same as `'adapt_full'`, but use uniform jitter in [-1, 1]
          as starting point in each chain.
    **kwargs: keyword arguments
        Extra keyword arguments are forwarded to littlemcmc.NUTS.

    Returns
    -------
    start: np.array
        Starting point for sampler.
    nuts_sampler: NUTS
        Instantiated and initialized NUTS sampler object.
    """
    if not isinstance(init, str):
        raise TypeError("init must be a string.")

    if init is not None:
        init = init.lower()

    if init == "auto":
        init = "jitter+adapt_diag"

    _log.info("Initializing NUTS using {}...".format(init))

    if random_seed is not None:
        random_seed = int(np.atleast_1d(random_seed)[0])
        np.random.seed(random_seed)

    if init == "adapt_diag":
        start = np.zeros(model_ndim)
        mean = start
        var = np.ones(model_ndim)
        potential: QuadPotential = QuadPotentialDiagAdapt(model_ndim, mean, var, 10)
    elif init == "jitter+adapt_diag":
        start = 2 * np.random.rand(model_ndim) - 1
        mean = start
        var = np.ones(model_ndim)
        potential = QuadPotentialDiagAdapt(model_ndim, mean, var, 10)
    elif init == "adapt_full":
        start = np.zeros(model_ndim)
        mean = start
        cov = np.eye(model_ndim)
        potential = QuadPotentialFullAdapt(model_ndim, mean, cov, 10)
    elif init == "jitter+adapt_full":
        start = 2 * np.random.rand(model_ndim) - 1
        mean = start
        cov = np.eye(model_ndim)
        potential = QuadPotentialFullAdapt(model_ndim, mean, cov, 10)
    else:
        raise ValueError("Unknown initializer: {}.".format(init))

    step = NUTS(
        logp_dlogp_func=logp_dlogp_func, model_ndim=model_ndim, potential=potential, **kwargs
    )

    return start, step



#  Copyright 2019-2020 George Ho
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import codecs
import re
from pathlib import Path
from setuptools import setup, find_packages


PROJECT_ROOT = Path(__file__).resolve().parent
REQUIREMENTS_FILE = PROJECT_ROOT / "requirements.txt"
README_FILE = PROJECT_ROOT / "README.md"
VERSION_FILE = PROJECT_ROOT / "littlemcmc" / "__init__.py"

NAME = "littlemcmc"
DESCRIPTION = "A lightweight and performant implementation of HMC and NUTS in Python, spun out of the PyMC project."
AUTHOR = "George Ho"
AUTHOR_EMAIL = "mail@eigenfoo.xyz"
URL = "https://github.com/eigenfoo/littlemcmc"
LICENSE = "Apache License, Version 2.0"

CLASSIFIERS = [
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "License :: OSI Approved :: Apache Software License",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Operating System :: OS Independent",
]


def get_requirements(path):
    with codecs.open(path) as buff:
        return buff.read().splitlines()


def get_long_description():
    with codecs.open(README_FILE, "rt") as buff:
        return buff.read()


def get_version():
    lines = open(VERSION_FILE, "rt").readlines()
    version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in lines:
        mo = re.search(version_regex, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError("Unable to find version in %s." % (VERSION_FILE,))


if __name__ == "__main__":
    setup(
        name=NAME,
        version=get_version(),
        description=DESCRIPTION,
        license=LICENSE,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        classifiers=CLASSIFIERS,
        packages=find_packages(),
        install_requires=get_requirements(REQUIREMENTS_FILE),
        long_description=get_long_description(),
        long_description_content_type="text/markdown",
        include_package_data=True,
    )



import numpy as np
from jax.config import config

config.update("jax_enable_x64", True)

import jax
import jax.numpy as jnp
import littlemcmc as lmc


np.random.seed(42)

true_params = np.array([0.5, -2.3, -0.23])

N = 50
t = np.linspace(0, 10, 2)
x = np.random.uniform(0, 10, 50)
y = x * true_params[0] + true_params[1]
y_obs = y + np.exp(true_params[-1]) * np.random.randn(N)


def jax_model(params):
    mean = params[0] * x + params[1]
    loglike = -0.5 * jnp.sum((y_obs - mean) ** 2 * jnp.exp(-2 * params[2]) + 2 * params[2])
    return loglike


@jax.jit
def jax_model_and_grad(x):
    return jax_model(x), jax.grad(jax_model)(x)


def jax_logp_dlogp_func(x):
    v, g = jax_model_and_grad(x)
    return np.asarray(v), np.asarray(g)


trace, stats = lmc.sample(
    logp_dlogp_func=jax_logp_dlogp_func, model_ndim=3, tune=500, draws=1000, chains=4,
)
